#
#
#   A class to write to the GraphViz format...
#
#

from neuromllite.utils import print_v
from neuromllite.utils import is_spiking_input_nml_cell
from neuromllite.ConnectivityHandler import ConnectivityHandler
from neuromllite.NetworkGenerator import _get_rng_for_network
from neuromllite.utils import evaluate

from pyneuroml.pynml import convert_to_units

from graphviz import Digraph
import numpy as np

engines = {
    "d": "dot",
    "c": "circo",
    "n": "neato",
    "t": "twopi",
    "f": "fdp",
    "s": "sfdp",
    "p": "patchwork",
}


class GraphVizHandler(ConnectivityHandler):

    DEFAULT_POP_SHAPE = "ellipse"
    EXC_POP_SHAPE = "ellipse"
    INH_POP_SHAPE = "ellipse"

    EXC_CONN_ARROW_SHAPE = "normal"
    INH_CONN_ARROW_SHAPE = "dot"
    GAP_CONN_ARROW_SHAPE = "tee"
    CONT_CONN_ARROW_SHAPE = "normal"
    INPUT_ARROW_SHAPE = "empty"

    def __init__(
        self,
        level=10,
        engine="dot",
        nl_network=None,
        include_ext_inputs=True,
        include_input_pops=True,
        scale_by_post_pop_size=True,
        scale_by_post_pop_cond=True,
        show_chem_conns=True,
        show_elect_conns=True,
        show_cont_conns=True,
        min_weight_to_show=0,
        output_format="png",
        view_on_render=True,
    ):

        self.nl_network = nl_network
        self.level = level
        self.engine = engine
        self.include_ext_inputs = include_ext_inputs
        self.include_input_pops = include_input_pops
        self.scale_by_post_pop_size = scale_by_post_pop_size
        self.scale_by_post_pop_cond = scale_by_post_pop_cond
        self.min_weight_to_show = min_weight_to_show
        self.show_chem_conns = show_chem_conns
        self.show_elect_conns = show_elect_conns
        self.show_cont_conns = show_cont_conns
        self.output_format = output_format
        self.view_on_render = view_on_render

        self.rng, seed = _get_rng_for_network(self.nl_network)

        print_v(
            "Initiating GraphViz handler, level %i, engine: %s, seed: %s"
            % (level, engine, seed)
        )

    def print_settings(self):
        print_v("**************************************")
        print_v("* Settings for GraphVizHandler: ")
        print_v("*")
        print_v("*    engine:                  %s" % self.engine)
        print_v("*    level:                   %s" % self.level)
        print_v("*    is_cell_level:           %s" % self.is_cell_level())
        print_v("*    CUTOFF_INH_SYN_MV:       %s" % self.CUTOFF_INH_SYN_MV)
        print_v("*    include_ext_inputs:      %s" % self.include_ext_inputs)
        print_v("*    include_input_pops:      %s" % self.include_input_pops)
        print_v("*    scale_by_post_pop_size:  %s" % self.scale_by_post_pop_size)
        print_v("*    scale_by_post_pop_cond:  %s" % self.scale_by_post_pop_cond)
        print_v("*    min_weight_to_show:      %s" % self.min_weight_to_show)
        print_v("*    show_chem_conns:         %s" % self.show_chem_conns)
        print_v("*    show_elect_conns:        %s" % self.show_elect_conns)
        print_v("*    show_cont_conns:         %s" % self.show_cont_conns)
        print_v("*    output_format:           %s" % self.output_format)
        print_v("*")
        print_v("* Used values: ")
        syns = sorted(self.syn_conds_used.keys())
        print_v(
            "*    syn_conds_used:          %s"
            % "\n*                             ".join(
                ["%s:\t %s" % (k, self.syn_conds_used[k]) for k in syns]
            )
        )
        print_v("*")
        print_v("**************************************")

    def get_weight_fraction_and_line(self, w, max_w, min_w):

        if max_w == min_w:
            return 1.0, 1.0
        # fractional weight, 0->1
        fweight = (abs(w) - min_w) / (max_w - min_w)
        if fweight < 0:
            fweight = 0
        if fweight > 1:
            fweight = 1

        # Good weight for line based on fraction
        lweight = 0.7 + fweight * 3.5

        return fweight, lweight

    def finalise_document(self):

        max_abs_weight = {}
        min_abs_weight = {}

        for t in [
            "projection",
            "inhibitory",
            "excitatory",
            "electricalProjection",
            "continuousProjection",
        ]:
            max_abs_weight[t] = -1e100
            min_abs_weight[t] = 1e100

        if self.is_cell_level() and self.level <= -1:

            for projName in self.proj_weights:
                # ws = self.proj_individual_weights[projName]
                ws = self.proj_individual_scaled_weights[projName]
                t = self.proj_types[projName]
                if np.max(ws) > 0:
                    max_abs_weight[t] = max(
                        max_abs_weight[t], np.max(ws[np.nonzero(ws)])
                    )
                    min_abs_weight[t] = min(
                        min_abs_weight[t], np.min(ws[np.nonzero(ws)])
                    )

            for projName in self.proj_weights:

                pre_pop = self.proj_pre_pops[projName]
                post_pop = self.proj_post_pops[projName]
                proj_type = self.proj_types[projName]
                num_pre = self.get_size_pre_pop(projName)
                num_post = self.get_size_post_pop(projName)

                if self.proj_types[projName] == "electricalProjection":
                    show = self.show_elect_conns
                elif self.proj_types[projName] == "continuousProjection":
                    show = self.show_cont_conns
                else:
                    show = self.show_chem_conns

                if show:
                    gbase_nS, gbase = self._get_gbase_nS(
                        projName, return_orig_string_also=True
                    )

                    pclass = self._get_proj_class(proj_type)
                    sign = -1 if "inhibitory" in proj_type else 1

                    print_v(
                        "GRAPH PROJ: %s (%s (%i) -> %s (%i), %s): w %s; wtot: %s; sign: %s; cond: %s nS (%s); all: %s -> %s"
                        % (
                            projName,
                            pre_pop,
                            num_pre,
                            post_pop,
                            num_post,
                            proj_type,
                            self.proj_weights[projName],
                            self.proj_tot_weight[projName],
                            sign,
                            gbase_nS,
                            gbase,
                            max_abs_weight,
                            min_abs_weight,
                        )
                    )

                    for pre_i in range(num_pre):
                        for post_i in range(num_post):
                            pre_pop_i = self.get_cell_identifier(pre_pop, pre_i)
                            post_pop_i = self.get_cell_identifier(post_pop, post_i)

                            w = self.proj_individual_scaled_weights[projName][pre_i][
                                post_i
                            ]
                            w_unscaled = self.proj_individual_weights[projName][pre_i][
                                post_i
                            ]

                            num_indiv_conns = self.proj_individual_conn_numbers[
                                projName
                            ][pre_i][post_i]

                            if w != 0:

                                weight_used = w

                                cond_scale = None
                                if self.scale_by_post_pop_cond:
                                    cond_scale = gbase_nS if gbase_nS != None else 1.0

                                print_v(
                                    " - conn %s -> %s: %s (%s)"
                                    % (pre_pop_i, post_pop_i, w, weight_used)
                                )

                                fweight, lweight = self.get_weight_fraction_and_line(
                                    weight_used,
                                    max_abs_weight[proj_type],
                                    min_abs_weight[proj_type],
                                )

                                self.graph.attr(
                                    "edge",
                                    style=self.proj_lines[projName],
                                    arrowhead=self.proj_shapes[projName],
                                    arrowsize="%s" % (min(1, lweight)),
                                    penwidth="%s" % (lweight),
                                    color=self.pop_colors[self.proj_pre_pops[projName]],
                                    fontcolor=self.pop_colors[
                                        self.proj_pre_pops[projName]
                                    ],
                                )

                                label = "<"
                                label += "w: %s <br/> " % self.format_float(w_unscaled)
                                label += "num: %i <br/> " % num_indiv_conns

                                if num_indiv_conns != 1:
                                    wc = float(w_unscaled) / num_indiv_conns
                                    label += "w/conn: %s <br/> " % self.format_float(wc)
                                    if w != w_unscaled:
                                        label += "%s*%s*%snS = %snS<br/> " % (
                                            self.format_float(wc),
                                            num_indiv_conns,
                                            self.format_float(cond_scale),
                                            self.format_float(weight_used),
                                        )
                                else:
                                    label += "%s*%snS = %snS<br/> " % (
                                        self.format_float(w),
                                        self.format_float(cond_scale),
                                        self.format_float(weight_used),
                                    )

                                label += "scaled: %s<br/> " % (
                                    self.format_float(weight_used)
                                )

                                label += "f: %s <br/> " % fweight
                                label += "l: %s <br/> " % lweight

                                delay = self.proj_delays[projName][pre_i][post_i]
                                if delay != 0:
                                    label += "d: %sms <br/> " % self.format_float(delay)

                                if not label[-1] == ">":
                                    label += ">"
                                if self.level <= -2:
                                    self.graph.edge(pre_pop_i, post_pop_i, label=label)
                                else:
                                    self.graph.edge(pre_pop_i, post_pop_i)

        else:

            for projName in self.proj_tot_weight:
                t = self.proj_types[projName]

                if self.proj_tot_weight[projName] != 0:

                    weight_used = float(self.proj_tot_weight[projName])
                    weight_used = self._scale_population_weight(weight_used, projName)

                    if abs(weight_used) >= self.min_weight_to_show:

                        max_abs_weight[t] = max(max_abs_weight[t], abs(weight_used))
                        min_abs_weight[t] = min(min_abs_weight[t], abs(weight_used))
                        print_v(
                            "EDGE: %s (%s): weight %s (all so far: %s -> %s)"
                            % (
                                projName,
                                t,
                                weight_used,
                                max_abs_weight[t],
                                min_abs_weight[t],
                            )
                        )

                    else:
                        print_v(
                            "IGNORING EDGE: %s: weight %s, less than %s (all used so far: %s -> %s)"
                            % (
                                projName,
                                weight_used,
                                self.min_weight_to_show,
                                max_abs_weight[t],
                                min_abs_weight[t],
                            )
                        )

            for projName in self.proj_tot_weight:

                proj_type = self.proj_types[projName]

                if self.proj_tot_weight[projName] != 0:

                    # weight_used = self.proj_weights[projName]
                    weight_used = float(self.proj_tot_weight[projName])
                    weight_used = self._scale_population_weight(weight_used, projName)

                    if abs(weight_used) >= self.min_weight_to_show:

                        if max_abs_weight[proj_type] == min_abs_weight[proj_type]:
                            fweight = 1
                            lweight = 1
                        elif self.proj_types[projName] == "electricalProjection":
                            # Dotted lines look bad when thick...
                            fweight = 1
                            lweight = 1
                        else:
                            fweight, lweight = self.get_weight_fraction_and_line(
                                weight_used,
                                max_abs_weight[proj_type],
                                min_abs_weight[proj_type],
                            )

                        if self.proj_types[projName] == "electricalProjection":
                            show = self.show_elect_conns
                        elif self.proj_types[projName] == "continuousProjection":
                            show = self.show_cont_conns
                        else:
                            show = self.show_chem_conns

                        pre_pop = self.proj_pre_pops[projName]
                        if pre_pop in self.pop_nml_component_objs:
                            if (
                                not self.include_input_pops
                                and is_spiking_input_nml_cell(
                                    self.pop_nml_component_objs[pre_pop]
                                )
                            ):
                                show = False

                        if self.level >= 2 and show:
                            print_v(
                                "EDGE: %s (%s): weight %s (all: %s -> %s); fw: %s; lw: %s"
                                % (
                                    projName,
                                    proj_type,
                                    weight_used,
                                    max_abs_weight[proj_type],
                                    min_abs_weight[proj_type],
                                    fweight,
                                    lweight,
                                )
                            )

                            self.graph.attr(
                                "edge",
                                style=self.proj_lines[projName],
                                arrowhead=self.proj_shapes[projName],
                                arrowsize="%s" % (min(1, lweight)),
                                penwidth="%s" % (lweight),
                                color=self.pop_colors[self.proj_pre_pops[projName]],
                                fontcolor=self.pop_colors[self.proj_pre_pops[projName]],
                            )

                            if self.level >= 4:
                                label = "<"

                                if self.level >= 5:
                                    if projName in self.proj_syn_objs:
                                        syn = self.proj_syn_objs[projName]
                                        label += "%s<br/> " % syn.id

                                if self.proj_tot_weight[projName] > 0:
                                    avg_weight = float(
                                        float(self.proj_tot_weight[projName])
                                        / self.proj_conns[projName]
                                    )
                                    label += "avg weight: %s<br/> " % self.format_float(
                                        avg_weight, approx=True
                                    )

                                if self.nl_network:
                                    proj = self.nl_network.get_child(
                                        projName, "projections"
                                    )
                                    if proj and proj.random_connectivity:
                                        label += (
                                            "p: %s<br/> "
                                            % proj.random_connectivity.probability
                                        )

                                if self.proj_conns[projName] > 0:
                                    label += (
                                        "%s conns<br/> " % self.proj_conns[projName]
                                    )

                                if self.level >= 5:

                                    if self.proj_conns[projName] > 0:

                                        pre_avg = (
                                            float(self.proj_conns[projName])
                                            / self.pop_sizes[
                                                self.proj_pre_pops[projName]
                                            ]
                                        )
                                        post_avg = (
                                            float(self.proj_conns[projName])
                                            / self.pop_sizes[
                                                self.proj_post_pops[projName]
                                            ]
                                        )

                                        label += " %s/pre &#8594; %s/post<br/> " % (
                                            self.format_float(pre_avg, approx=True),
                                            self.format_float(post_avg, approx=True),
                                        )

                                        gbase_nS, gbase = self._get_gbase_nS(
                                            projName, return_orig_string_also=True
                                        )

                                        if gbase_nS != None:
                                            label += "%s*%s*%s = %s nS<br/> " % (
                                                self.format_float(post_avg),
                                                self.format_float(avg_weight),
                                                gbase,
                                                self.format_float(
                                                    post_avg * avg_weight * gbase_nS
                                                ),
                                            )

                                if self.level >= 6:
                                    label += "scaled weight: %s<br/> " % (
                                        self.format_float(weight_used)
                                    )
                                    label += "fract: %s<br/> " % (
                                        self.format_float(fweight)
                                    )
                                    label += "line: %s<br/> " % (lweight)

                                if not label[-1] == ">":
                                    label += ">"

                                self.graph.edge(
                                    self.proj_pre_pops[projName],
                                    self.proj_post_pops[projName],
                                    label=label,
                                )
                            else:
                                self.graph.edge(
                                    self.proj_pre_pops[projName],
                                    self.proj_post_pops[projName],
                                )

        print_v("Generating graph for: %s" % self.network_id)
        self.print_settings()

        if self.view_on_render:
            self.graph.view()
        else:
            self.graph.render()

        if self.nl_network:
            print_v(
                "Finished generating graph with params: %s"
                % [p for p in self.nl_network.parameters]
                if self.nl_network.parameters is not None
                else "[]"
            )

    def handle_network(self, network_id, notes, temperature=None):

        print_v("Network: %s" % network_id)
        self.network_id = network_id

        self.graph = Digraph(
            network_id,
            filename="%s.gv" % network_id,
            engine=self.engine,
            format=self.output_format,
        )

    def handle_population(
        self,
        population_id,
        component,
        size=-1,
        component_obj=None,
        properties={},
        notes=None,
    ):
        sizeInfo = " as yet unspecified size"

        if size >= 0:
            sizeInfo = ", size: " + str(size) + " cells"

        if component_obj:
            compInfo = " (%s)" % component_obj.__class__.__name__
        else:
            compInfo = ""

        self.pop_nml_component_objs[population_id] = component_obj

        if not self.include_input_pops and is_spiking_input_nml_cell(component_obj):
            print("Ignoring %s as it's a spiking input population")
            return

        self.pop_sizes[population_id] = size

        print_v(
            "Population: "
            + population_id
            + ", component: "
            + component
            + compInfo
            + sizeInfo
            + ", properties: %s" % properties
        )
        color = "#444444"
        fcolor = "#ffffff"
        shape = self.DEFAULT_POP_SHAPE

        if properties and "color" in properties:
            rgb = properties["color"].split()
            color = "#"
            for a in rgb:
                color = color + "%02x" % int(float(a) * 255)

            # https://stackoverflow.com/questions/3942878
            if (
                float(rgb[0]) * 0.299 + float(rgb[1]) * 0.587 + float(rgb[2]) * 0.2
            ) > 0.25:
                fcolor = "#000000"
            else:
                fcolor = "#ffffff"

            # print('Color %s -> %s -> %s'%(properties['color'], rgb, color))

        if properties and "type" in properties:
            self.pop_types[population_id] = properties["type"]
            if properties["type"] == "E":
                shape = self.EXC_POP_SHAPE
            if properties["type"] == "I":
                shape = self.INH_POP_SHAPE

        self.pop_colors[population_id] = color

        label = "<%s" % population_id
        if self.level >= 3:
            label += "<br/><i>%s cell%s</i>" % (size, "" if size == 1 else "s")
        if self.level >= 4:

            from neuroml import SpikeSourcePoisson

            if component_obj and isinstance(component_obj, SpikeSourcePoisson):
                start = convert_to_units(component_obj.start, "ms")
                if start == int(start):
                    start = int(start)
                duration = convert_to_units(component_obj.duration, "ms")
                if duration == int(duration):
                    duration = int(duration)
                rate = convert_to_units(component_obj.rate, "Hz")
                if rate == int(rate):
                    rate = int(rate)

                label += "<br/>Spikes %s-%sms@%sHz" % (start, start + duration, rate)

            else:
                label += "<br/>%s" % (component)

        label += ">"

        if properties and "region" in properties:

            with self.graph.subgraph(name="cluster_%s" % properties["region"]) as c:
                c.attr(color="#444444", fontcolor="#444444")
                c.attr(label=properties["region"])
                c.attr(
                    "node", color=color, style="filled", fontcolor=fcolor, shape=shape
                )

                if self.is_cell_level():
                    for i in range(size):
                        cell_info = self.get_cell_identifier(population_id, i)
                        c.node(cell_info, label=cell_info)
                else:
                    c.node(population_id, label=label)

        else:
            self.graph.attr(
                "node", color=color, style="filled", fontcolor=fcolor, shape=shape
            )

            if self.is_cell_level():
                for i in range(size):
                    cell_info = self.get_cell_identifier(population_id, i)
                    self.graph.node(cell_info, label=cell_info)
            else:
                self.graph.node(population_id, label=label)

    def handle_projection(
        self,
        projName,
        prePop,
        postPop,
        synapse,
        hasWeights=False,
        hasDelays=False,
        type="projection",
        synapse_obj=None,
        pre_synapse_obj=None,
    ):

        shape = self.EXC_CONN_ARROW_SHAPE
        line = "solid"

        # Could be used in a network with no explicit connections, but weight
        # between populations set at high level projection element in NeuroMLlite
        projection_weight = 1.0

        if not (prePop in self.pop_sizes and postPop in self.pop_sizes):
            print_v("Ignoring projection, as one of pops empty...")
            return

        self.proj_pre_pops[projName] = prePop
        self.proj_post_pops[projName] = postPop
        self.proj_types[projName] = type

        if prePop in self.pop_types:
            if "I" in self.pop_types[prePop]:
                shape = self.INH_CONN_ARROW_SHAPE

        if type == "electricalProjection":
            shape = self.GAP_CONN_ARROW_SHAPE
            line = "dashed"

        if type == "continuousProjection":
            shape = self.CONT_CONN_ARROW_SHAPE
            line = "solid"

        if synapse_obj:
            self.proj_syn_objs[projName] = synapse_obj
            erev = self.get_reversal_potential_mV(synapse_obj)
            if erev != None and erev < self.CUTOFF_INH_SYN_MV:
                shape = self.INH_CONN_ARROW_SHAPE

        if self.nl_network:
            syn = self.nl_network.get_child(synapse, "synapses")
            if syn:
                if syn.parameters:
                    if (
                        "e_rev" in syn.parameters
                        and syn.parameters["e_rev"] < self.CUTOFF_INH_SYN_MV
                    ):
                        shape = self.INH_CONN_ARROW_SHAPE

            proj = self.nl_network.get_child(projName, "projections")
            if proj:
                if proj.weight:
                    proj_weight = evaluate(
                        proj.weight, self.nl_network.parameters, self.rng
                    )
                    if proj_weight < 0:
                        shape = self.INH_CONN_ARROW_SHAPE
                    projection_weight = abs(proj_weight)

        self.proj_weights[projName] = projection_weight
        self.proj_shapes[projName] = shape
        self.proj_lines[projName] = line
        self.proj_conns[projName] = 0
        self.proj_tot_weight[projName] = 0

        if self.is_cell_level():
            pre_size = self.pop_sizes[prePop]
            post_size = self.pop_sizes[postPop]
            self.proj_individual_weights[projName] = np.zeros((pre_size, post_size))
            self.proj_individual_conn_numbers[projName] = np.zeros(
                (pre_size, post_size)
            )
            self.proj_individual_scaled_weights[projName] = np.zeros(
                (pre_size, post_size)
            )
            self.proj_delays[projName] = np.zeros((pre_size, post_size))

    def finalise_projection(
        self, projName, prePop, postPop, synapse=None, type="projection"
    ):

        # weight = self.proj_tot_weight[projName]
        # self.max_weight = max(self.max_weight, weight)
        # self.min_weight = min(self.min_weight, weight)
        # print_v("Now weights range %s->%s"%(self.min_weight, self.max_weight))
        pass
        # print_v("Projection finalising: "+projName+" from "+prePop+" to "+postPop+" completed")

    def finalise_input_source(self, inputListId):

        if self.include_ext_inputs:
            # self.print_input_information('FINAL: '+inputListId, self.pops_ils[inputListId], '...', self.sizes_ils[inputListId])

            if self.level >= 2:

                label = "<%s" % inputListId
                if self.level >= 3:
                    size = self.sizes_ils[inputListId]
                    label += "<br/><i>%s input%s</i>" % (size, "" if size == 1 else "s")

                if self.level >= 4:

                    from neuroml import PulseGenerator
                    from neuroml import TransientPoissonFiringSynapse
                    from neuroml import PoissonFiringSynapse
                    from pyneuroml.pynml import convert_to_units

                    input_comp_obj = self.input_comp_obj_ils[inputListId]

                    if input_comp_obj and isinstance(input_comp_obj, PulseGenerator):
                        start = convert_to_units(input_comp_obj.delay, "ms")
                        if start == int(start):
                            start = int(start)
                        duration = convert_to_units(input_comp_obj.duration, "ms")
                        if duration == int(duration):
                            duration = int(duration)
                        amplitude = convert_to_units(input_comp_obj.amplitude, "pA")
                        if amplitude == int(amplitude):
                            amplitude = int(amplitude)

                        label += "<br/>Pulse %s-%sms@%spA" % (
                            start,
                            start + duration,
                            amplitude,
                        )

                    if input_comp_obj and isinstance(
                        input_comp_obj, PoissonFiringSynapse
                    ):

                        average_rate = convert_to_units(
                            input_comp_obj.average_rate, "Hz"
                        )
                        if average_rate == int(average_rate):
                            average_rate = int(average_rate)

                        label += "<br/>Syn: %s @ %sHz" % (
                            input_comp_obj.synapse,
                            average_rate,
                        )

                    if input_comp_obj and isinstance(
                        input_comp_obj, TransientPoissonFiringSynapse
                    ):

                        start = convert_to_units(input_comp_obj.delay, "ms")
                        if start == int(start):
                            start = int(start)
                        duration = convert_to_units(input_comp_obj.duration, "ms")
                        if duration == int(duration):
                            duration = int(duration)
                        average_rate = convert_to_units(
                            input_comp_obj.average_rate, "Hz"
                        )
                        if average_rate == int(average_rate):
                            average_rate = int(average_rate)

                        label += "<br/>Syn: %s<br/>%s-%sms @ %sHz" % (
                            input_comp_obj.synapse,
                            start,
                            start + duration,
                            average_rate,
                        )

                label += ">"

                self.graph.attr("node", color="#444444", style="", fontcolor="#444444")
                self.graph.node(inputListId, label=label)

                label = None
                if self.level >= 5:
                    label = "<"
                    if self.sizes_ils[inputListId] > 0:
                        percent = (
                            100
                            * float(self.sizes_ils[inputListId])
                            / self.pop_sizes[self.pops_ils[inputListId]]
                        )

                        if percent <= 100:
                            label += "%s%s%% of population<br/> " % (
                                " " if percent != 100 else "",
                                self.format_float(percent),
                            )
                        else:
                            label += "%s%s per cell<br/> " % (
                                " ",
                                self.format_float(percent / 100),
                            )

                        avg_weight = (
                            float(self.weights_ils[inputListId])
                            / self.sizes_ils[inputListId]
                        )
                        label += "avg weight: %s<br/> " % (
                            self.format_float(avg_weight, approx=True)
                        )

                    if not label[-1] == ">":
                        label += ">"

                self.graph.edge(
                    inputListId,
                    self.pops_ils[inputListId],
                    arrowhead=self.INPUT_ARROW_SHAPE,
                    label=label,
                )



from neuromllite import Network, Cell, InputSource, Population, Synapse
from neuromllite import Projection, RandomConnectivity, Input, Simulation
import sys

################################################################################
###   Build new network

net = Network(id="Spikers")
net.notes = "Example with spiking entities.."
net.parameters = {"N": 10, "weightInput": 10, "input_rate": 40}

cell = Cell(id="iafcell", pynn_cell="IF_cond_alpha")
cell.parameters = {"tau_refrac": 10}
net.cells.append(cell)

input_cell = Cell(id="InputCell", pynn_cell="SpikeSourcePoisson")
input_cell.parameters = {"start": 0, "duration": 10000000000, "rate": "input_rate"}
net.cells.append(input_cell)

input_cell_100 = Cell(id="InputCell100", pynn_cell="SpikeSourcePoisson")
input_cell_100.parameters = {"start": 0, "duration": 10000000000, "rate": 100}
net.cells.append(input_cell_100)


input_source_p0 = InputSource(
    id="poissonFiringSyn", neuroml2_source_file="../test_files/inputs.nml"
)
net.input_sources.append(input_source_p0)

input_source1 = InputSource(
    id="iclamp1",
    pynn_input="DCSource",
    parameters={"amplitude": 0.8, "start": 100.0, "stop": 900.0},
)

net.input_sources.append(input_source1)


pop0 = Population(
    id="pop0", size="N", component=cell.id, properties={"color": ".7 0 0"}
)
net.populations.append(pop0)
pop1 = Population(
    id="pop1", size="N", component=cell.id, properties={"color": "0 .7 0"}
)
net.populations.append(pop1)
pop2 = Population(
    id="pop2", size="N", component=cell.id, properties={"color": "0 .7 0.7"}
)
net.populations.append(pop2)
#

ipop0 = Population(
    id="input_pop0", size="N", component=input_cell.id, properties={"color": ".7 .7 .7"}
)
net.populations.append(ipop0)
ipop1 = Population(
    id="input_pop1",
    size="N",
    component=input_cell_100.id,
    properties={"color": ".7 .1 .7"},
)
net.populations.append(ipop1)

# pRS = Population(id='RSpop', size='N - int(N*fractionE)', component=cell.id, properties={'color':'0 0 .7'})

# net.populations.append(pRS)
"""
net.synapses.append(Synapse(id='ampa', neuroml2_source_file='test_files/ampa.synapse.nml'))
#net.synapses.append(Synapse(id='gaba', neuroml2_source_file='test_files/gaba.synapse.nml'))


net.projections.append(Projection(id='projEI',
                                  presynaptic=pE.id,
                                  postsynaptic=pRS.id,
                                  synapse='ampa',
                                  delay=2,
                                  weight=0.2,
                                  random_connectivity=RandomConnectivity(probability=.8)))

                            """
net.inputs.append(
    Input(
        id="stim0",
        input_source=input_source_p0.id,
        population=pop0.id,
        percentage=50,
        weight="weightInput",
    )
)

net.inputs.append(
    Input(id="stim1", input_source=input_source1.id, population=pop1.id, percentage=100)
)

net.inputs.append(
    Input(id="stim2", input_source=input_source1.id, population=pop2.id, percentage=50)
)

print(net)
print(net.to_json())
new_file = net.to_json_file("%s.json" % net.id)


################################################################################
###   Build Simulation object & save as JSON

sim = Simulation(
    id="SimSpikers",
    network=new_file,
    duration="10000",
    dt="0.025",
    record_traces={"pop0": "*", "pop1": "*", "pop2": "*"},
    record_spikes={"all": "*"},
)

sim.to_json_file()


################################################################################
###   Run in some simulators

from neuromllite.NetworkGenerator import check_to_generate_or_run
import sys

check_to_generate_or_run(sys.argv, sim)



from neuromllite import (
    Network,
    Cell,
    InputSource,
    Population,
    Synapse,
    RectangularRegion,
    RandomLayout,
)
from neuromllite import Projection, RandomConnectivity, Input, Simulation
import sys

################################################################################
###   Build new network

net = Network(id="Example11_Synapses")
net.notes = "Example 11: synaptic properties"
net.parameters = {"input_amp": 0.23, "weight": 1.01}
#'tau_syn':     2}

cell = Cell(id="iafCell0", neuroml2_source_file="test_files/iaf.cell.nml")
# cell.parameters = { "tau_refrac":5, "i_offset":0 }
net.cells.append(cell)


input_source = InputSource(
    id="i_clamp",
    pynn_input="DCSource",
    parameters={"amplitude": "input_amp", "start": 200.0, "stop": 800.0},
)
net.input_sources.append(input_source)

r1 = RectangularRegion(id="region1", x=0, y=0, z=0, width=1000, height=100, depth=1000)
net.regions.append(r1)

p0 = Population(
    id="pop0",
    size=1,
    component=cell.id,
    properties={"color": "1 0 0"},
    random_layout=RandomLayout(region=r1.id),
)
p1 = Population(
    id="pop1",
    size=1,
    component=cell.id,
    properties={"color": "0 1 0"},
    random_layout=RandomLayout(region=r1.id),
)

net.populations.append(p0)
net.populations.append(p1)

ampaSyn = Synapse(id="ampa", neuroml2_source_file="test_files/ampa.synapse.nml")
net.synapses.append(ampaSyn)

nmdaSyn = Synapse(id="nmdaSyn", neuroml2_source_file="test_files/NMDA.synapse.nml")
net.synapses.append(nmdaSyn)

net.projections.append(
    Projection(
        id="proj0",
        presynaptic=p0.id,
        postsynaptic=p1.id,
        synapse=ampaSyn.id,
        delay=0,
        weight="weight",
    )
)
net.projections[0].random_connectivity = RandomConnectivity(probability=1)

net.projections.append(
    Projection(
        id="proj1",
        presynaptic=p0.id,
        postsynaptic=p1.id,
        synapse=nmdaSyn.id,
        delay=0,
        weight="weight",
    )
)
net.projections[1].random_connectivity = RandomConnectivity(probability=1)


net.inputs.append(
    Input(id="stim", input_source=input_source.id, population=p0.id, percentage=100)
)

print(net.to_json())
new_file = net.to_json_file("%s.json" % net.id)


################################################################################
###   Build Simulation object & save as JSON

sim = Simulation(
    id="SimExample11",
    network=new_file,
    duration="1000",
    dt="0.01",
    record_traces={"all": "*"},
    record_variables={
        "synapses:%s:0/g" % ampaSyn.id: {"pop1": "*"},
        "synapses:%s:0/g" % nmdaSyn.id: {"pop1": "*"},
    },
    record_spikes={"pop0": "*"},
)

sim.to_json_file()


################################################################################
###   Run in some simulators

from neuromllite.NetworkGenerator import check_to_generate_or_run
import sys

check_to_generate_or_run(sys.argv, sim)



#
#
#   A class to write the network connectivity in matrix format...
#
#

from neuromllite.utils import print_v
from neuromllite.ConnectivityHandler import ConnectivityHandler

from neuromllite.utils import evaluate
from neuromllite.NetworkGenerator import _get_rng_for_network

import numpy as np


class MatrixHandler(ConnectivityHandler):

    colormaps_used = []

    weight_arrays_to_show = {}

    def __init__(self, level=10, nl_network=None):

        self.nl_network = nl_network
        self.level = level

        self.rng, seed = _get_rng_for_network(self.nl_network)

        print_v("Initiating Matrix handler, level %i, seed: %s" % (level, seed))
        self.scale_by_post_pop_cond = True

    def print_settings(self):
        print_v("**************************************")
        print_v("*")
        print_v("* Settings for MatrixHandler: ")
        print_v("*    level:                   %s" % self.level)
        print_v("*    is_cell_level:           %s" % self.is_cell_level())
        print_v("*    CUTOFF_INH_SYN_MV:       %s" % self.CUTOFF_INH_SYN_MV)
        # print_v('*    include_inputs:          %s'%self.include_inputs)
        # print_v('*    scale_by_post_pop_size:  %s'%self.scale_by_post_pop_size)
        # print_v('*    scale_by_post_pop_cond:  %s'%self.scale_by_post_pop_cond)
        # print_v('*    min_weight_to_show:      %s'%self.min_weight_to_show)
        # print_v('*    min_weight_to_show:      %s'%self.min_weight_to_show)
        print_v("*")
        print_v("* Used values: ")
        print_v("*    colormaps_used:          %s" % self.colormaps_used)
        print_v("*    zero_weight_color:       %s" % self.zero_weight_color)
        syns = sorted(self.syn_conds_used.keys())
        print_v(
            "*    syn_conds_used:          %s"
            % "\n*                             ".join(
                ["%s:\t %s" % (k, self.syn_conds_used[k]) for k in syns]
            )
        )
        print_v("*")
        print_v("**************************************")

    def _get_conn_label(self, template, pclass):
        p = pclass[0].upper() + pclass[1:]
        p = p.replace("_", " ")
        label = template % ("%s conns" % p)
        return label

    def finalise_document(self):

        entries = []
        # print_v('Finals: %s -> %s'%(self.proj_pre_pops, self.proj_post_pops))
        all_pops = []
        for v in self.proj_pre_pops.values():
            all_pops.append(v)
        for v in self.proj_post_pops.values():
            all_pops.append(v)

        for pop in all_pops:
            if self.is_cell_level():
                for i in range(self.pop_sizes[pop]):
                    pi = self.get_cell_identifier(pop, i)  # '%s_%i'%(pop,i)
                    if not pi in entries:
                        entries.append(pi)
            else:
                if not pop in entries:
                    entries.append(pop)

        entries = sorted(entries)

        matrix_per_cell = "%s (total weight between cell pair)"
        matrix_per_cell_cond = "%s (total conn weight*syn cond)"
        matrix_per_cell_cond_signed = "%s (total signed conn weight*syn cond)"

        matrix_single_conns = "%s (avg weight per existant conn)"
        matrix_number_conns = "%s (number of conns)"
        matrix_total_conns = "%s (sum of signed weights of conns)"
        matrix_total_conns_per_cell = "%s ((sum of weights)*cond/postpopsize)"
        matrix_total_signed_conns_per_cell = (
            "%s ((sum of signed weights)*cond/postpopsize)"
        )

        cbar_labels = {}
        for proj_type in self.proj_types.values():
            pclass = self._get_proj_class(proj_type)

            all_matrices = [
                matrix_single_conns,
                matrix_number_conns,
                matrix_total_conns,
                matrix_total_conns_per_cell,
                matrix_total_signed_conns_per_cell,
            ]

            if self.is_cell_level():
                all_matrices = [
                    matrix_per_cell,
                    matrix_per_cell_cond,
                    matrix_per_cell_cond_signed,
                ]

            for m in all_matrices:
                label = self._get_conn_label(m, pclass)

                self.weight_arrays_to_show[label] = np.zeros(
                    (len(entries), len(entries))
                )

            cbar_labels[self._get_conn_label(matrix_per_cell, pclass)] = "weight"
            cbar_labels[
                self._get_conn_label(matrix_per_cell_cond, pclass)
            ] = "conductance (nS)"
            cbar_labels[
                self._get_conn_label(matrix_per_cell_cond_signed, pclass)
            ] = "conductance * sign (nS)"

            cbar_labels[self._get_conn_label(matrix_number_conns, pclass)] = "number"
            cbar_labels[
                self._get_conn_label(matrix_single_conns, pclass)
            ] = "weight"  # (red: exc; blue: inh)'
            cbar_labels[
                self._get_conn_label(matrix_total_conns, pclass)
            ] = "total weight"  # (red: exc; blue: inh)'
            cbar_labels[
                self._get_conn_label(matrix_total_conns_per_cell, pclass)
            ] = "conductance (nS)"
            cbar_labels[
                self._get_conn_label(matrix_total_signed_conns_per_cell, pclass)
            ] = "conductance * sign (nS)"

        for projName in self.proj_weights:

            pre_pop = self.proj_pre_pops[projName]
            post_pop = self.proj_post_pops[projName]
            proj_type = self.proj_types[projName]
            num_pre = self.get_size_pre_pop(projName)
            num_post = self.get_size_post_pop(projName)

            gbase_nS, gbase = self._get_gbase_nS(projName, return_orig_string_also=True)

            pclass = self._get_proj_class(proj_type)
            sign = -1 if "inhibitory" in proj_type else 1

            print_v(
                "MATRIX PROJ: %s (%s (%i) -> %s (%i), %s): w %s; wtot: %s; sign: %s; cond: %s nS (%s)"
                % (
                    projName,
                    pre_pop,
                    num_pre,
                    post_pop,
                    num_post,
                    proj_type,
                    self.proj_weights[projName],
                    self.proj_tot_weight[projName],
                    sign,
                    gbase_nS,
                    gbase,
                )
            )

            if self.is_cell_level():

                for pre_i in range(self.pop_sizes[pre_pop]):
                    for post_i in range(self.pop_sizes[post_pop]):
                        pre_pop_i = entries.index(
                            self.get_cell_identifier(pre_pop, pre_i)
                        )
                        post_pop_i = entries.index(
                            self.get_cell_identifier(post_pop, post_i)
                        )

                        self.weight_arrays_to_show[
                            self._get_conn_label(matrix_per_cell, pclass)
                        ][pre_pop_i][post_pop_i] += self.proj_individual_weights[
                            projName
                        ][
                            pre_i
                        ][
                            post_i
                        ]
                        if projName in self.proj_syn_objs:
                            w_scaled = self.proj_individual_scaled_weights[projName][
                                pre_i
                            ][post_i]
                            self.weight_arrays_to_show[
                                self._get_conn_label(matrix_per_cell_cond, pclass)
                            ][pre_pop_i][post_pop_i] += w_scaled
                            self.weight_arrays_to_show[
                                self._get_conn_label(
                                    matrix_per_cell_cond_signed, pclass
                                )
                            ][pre_pop_i][post_pop_i] += (w_scaled * sign)

            else:
                pre_pop_i = entries.index(pre_pop)
                post_pop_i = entries.index(post_pop)

                self.weight_arrays_to_show[
                    self._get_conn_label(matrix_total_conns, pclass)
                ][pre_pop_i][post_pop_i] += (abs(self.proj_tot_weight[projName]) * sign)

                if abs(self.proj_tot_weight[projName]) != abs(
                    self.proj_weights[projName]
                ):

                    self.weight_arrays_to_show[
                        self._get_conn_label(matrix_single_conns, pclass)
                    ][pre_pop_i][post_pop_i] += (
                        abs(self.proj_weights[projName]) * sign
                    )

                    self.weight_arrays_to_show[
                        self._get_conn_label(matrix_number_conns, pclass)
                    ][pre_pop_i][post_pop_i] += abs(self.proj_conns[projName])

                    cond_scale = gbase_nS if gbase_nS != None else 1.0
                    tot_scaled = (
                        abs(self.proj_tot_weight[projName]) * cond_scale / num_post
                    )
                    self.weight_arrays_to_show[
                        self._get_conn_label(matrix_total_conns_per_cell, pclass)
                    ][pre_pop_i][post_pop_i] += tot_scaled
                    self.weight_arrays_to_show[
                        self._get_conn_label(matrix_total_signed_conns_per_cell, pclass)
                    ][pre_pop_i][post_pop_i] += (sign * tot_scaled)

        import matplotlib.pyplot as plt
        import matplotlib

        for proj_type in self.weight_arrays_to_show:
            weight_array = self.weight_arrays_to_show[proj_type]

            print_v(
                "Plotting proj_type: %s with values from %s to %s"
                % (proj_type, weight_array.min(), weight_array.max())
            )
            if not (weight_array.max() == 0 and weight_array.min() == 0):

                fig, ax = plt.subplots()
                title = "%s" % (proj_type)
                title2 = "%s" % (proj_type)
                plt.title(title)
                fig.canvas.set_window_title(title2)

                max_abs_weight = max(weight_array.max(), -1.0 * (weight_array.min()))
                min_abs_weight = np.min(abs(weight_array[np.nonzero(weight_array)]))

                if weight_array.min() < 0:
                    cm = matplotlib.cm.get_cmap("bwr")
                    self.zero_weight_color = "green"

                else:
                    # cm = matplotlib.cm.get_cmap('binary')

                    # if 'indiv' in proj_type or 'number' in proj_type:
                    cm = matplotlib.cm.get_cmap("rainbow")
                    self.zero_weight_color = "black"

                if not cm.name in self.colormaps_used:
                    self.colormaps_used.append(str(cm.name))

                print_v(
                    "  Plotting weight matrix [%s] (%s; 0=%s) with vals %s -> %s (max abs: %s, min nz abs: %s)"
                    % (
                        title,
                        cm.name,
                        self.zero_weight_color,
                        weight_array.min(),
                        weight_array.max(),
                        max_abs_weight,
                        min_abs_weight,
                    )
                )

                im = plt.imshow(
                    weight_array, cmap=cm, interpolation="nearest", norm=None
                )

                if weight_array.min() < 0:
                    plt.clim(-1 * max_abs_weight, max_abs_weight)
                elif min_abs_weight == max_abs_weight:
                    plt.clim(max_abs_weight * 0.000001, max_abs_weight)
                else:
                    plt.clim(min_abs_weight * 0.9999, max_abs_weight)

                cm.set_under(self.zero_weight_color)

                ax = plt.gca()

                # Gridlines based on minor ticks
                if weight_array.shape[0] < 40:
                    ax.grid(which="minor", color="grey", linestyle="-", linewidth=0.3)

                xt = np.arange(weight_array.shape[1]) + 0
                ax.set_xticks(xt)
                ax.set_xticks(xt[:-1] + 0.5, minor=True)
                ax.set_yticks(np.arange(weight_array.shape[0]) + 0)
                ax.set_yticks(np.arange(weight_array.shape[0]) + 0.5, minor=True)

                ax.set_yticklabels(entries)
                ax.set_xticklabels(entries)
                ax.set_ylabel("presynaptic")
                tick_size = (
                    10
                    if weight_array.shape[0] < 20
                    else (8 if weight_array.shape[0] < 40 else 6)
                )
                ax.tick_params(axis="y", labelsize=tick_size)
                ax.set_xlabel("postsynaptic")
                ax.tick_params(axis="x", labelsize=tick_size)
                fig.autofmt_xdate()

                for i in range(len(entries)):
                    alpha = 1
                    lwidth = 7
                    offset = -1 * lwidth * len(entries) / 500.0

                    if self.pop_colors[entries[i]]:
                        from matplotlib import lines

                        x, y = [[-0.5 + offset, -0.5 + offset], [i - 0.5, i + 0.5]]
                        line = lines.Line2D(
                            x,
                            y,
                            lw=lwidth,
                            color=self.pop_colors[entries[i]],
                            alpha=alpha,
                        )
                        line.set_solid_capstyle("butt")
                        line.set_clip_on(False)
                        ax.add_line(line)

                        x, y = [
                            [i - 0.5, i + 0.5],
                            [len(entries) - 0.5 - offset, len(entries) - 0.5 - offset],
                        ]
                        line = lines.Line2D(
                            x,
                            y,
                            lw=lwidth,
                            color=self.pop_colors[entries[i]],
                            alpha=alpha,
                        )
                        line.set_solid_capstyle("butt")
                        line.set_clip_on(False)
                        ax.add_line(line)

                cbar = plt.colorbar(im)
                if proj_type in cbar_labels:
                    cbar.set_label(cbar_labels[proj_type])

        print_v("Generating matrix for: %s" % self.network_id)
        self.print_settings()

        plt.show()

    def handle_population(
        self,
        population_id,
        component,
        size=-1,
        component_obj=None,
        properties={},
        notes=None,
    ):
        sizeInfo = " as yet unspecified size"

        if size >= 0:
            sizeInfo = ", size: " + str(size) + " cells"

        if component_obj:
            compInfo = " (%s)" % component_obj.__class__.__name__
        else:
            compInfo = ""

        print_v(
            "Population: "
            + population_id
            + ", component: "
            + component
            + compInfo
            + sizeInfo
            + ", properties: %s" % properties
        )

        color = None

        if properties and "color" in properties:
            rgb = properties["color"].split()
            color = "#"
            for a in rgb:
                color = color + "%02x" % int(float(a) * 255)

            # https://stackoverflow.com/questions/3942878
            if (
                float(rgb[0]) * 0.299 + float(rgb[1]) * 0.587 + float(rgb[2]) * 0.2
            ) > 0.25:
                fcolor = "#000000"
            else:
                fcolor = "#ffffff"

            print_v("Color %s -> %s -> %s" % (properties["color"], rgb, color))

        pop_type = None
        if properties and "type" in properties:
            pop_type = properties["type"]

        self.pop_sizes[population_id] = size

        if not self.is_cell_level():
            self.pop_colors[population_id] = color
            if pop_type:
                self.pop_types[population_id] = pop_type
        else:
            for i in range(size):
                cell_pop = self.get_cell_identifier(population_id, i)
                self.pop_colors[cell_pop] = color
                if pop_type:
                    self.pop_types[cell_pop] = pop_type

    def handle_projection(
        self,
        projName,
        prePop,
        postPop,
        synapse,
        hasWeights=False,
        hasDelays=False,
        type="projection",
        synapse_obj=None,
        pre_synapse_obj=None,
    ):

        weight = 1.0
        self.proj_pre_pops[projName] = prePop
        self.proj_post_pops[projName] = postPop

        proj_type = "excitatory"

        if prePop in self.pop_types:
            if "I" in self.pop_types[prePop]:
                proj_type = "inhibitory"

        if type == "electricalProjection":
            proj_type = "gap_junction"

        if synapse_obj:
            self.proj_syn_objs[projName] = synapse_obj
            erev = self.get_reversal_potential_mV(synapse_obj)
            if erev != None and erev < self.CUTOFF_INH_SYN_MV:
                proj_type = "inhibitory"

        if self.nl_network:
            syn = self.nl_network.get_child(synapse, "synapses")
            if syn:
                if syn.parameters:
                    if (
                        "e_rev" in syn.parameters
                        and syn.parameters["e_rev"] < self.CUTOFF_INH_SYN_MV
                    ):
                        proj_type = "inhibitory"

            proj = self.nl_network.get_child(projName, "projections")
            if proj:
                if proj.weight:
                    proj_weight = evaluate(
                        proj.weight, self.nl_network.parameters, self.rng
                    )
                    if proj_weight < 0:
                        proj_type = "inhibitory"
                    weight = float(abs(proj_weight))

        if type == "continuousProjection":
            proj_type += "continuous"

        self.proj_weights[projName] = float(weight)
        self.proj_types[projName] = proj_type
        self.proj_conns[projName] = 0
        self.proj_tot_weight[projName] = 0

        if self.is_cell_level():
            pre_size = self.pop_sizes[prePop]
            post_size = self.pop_sizes[postPop]
            self.proj_individual_weights[projName] = np.zeros((pre_size, post_size))
            self.proj_individual_conn_numbers[projName] = np.zeros(
                (pre_size, post_size)
            )
            self.proj_individual_scaled_weights[projName] = np.zeros(
                (pre_size, post_size)
            )
            self.proj_delays[projName] = np.zeros((pre_size, post_size))

        print_v(
            "New projection: %s, %s->%s, weights? %s, type: %s"
            % (projName, prePop, postPop, weight, proj_type)
        )

    def finalise_projection(
        self, projName, prePop, postPop, synapse=None, type="projection"
    ):

        pass
        # print_v("Projection finalising: "+projName+" -> "+prePop+" to "+postPop+" completed (%s; w: %s, conns: %s, tot w: %s)" % \
        #           (self.proj_types[projName], self.proj_weights[projName], self.proj_conns[projName], self.proj_tot_weight[projName]))

    sizes_ils = {}
    pops_ils = {}
    weights_ils = {}
    input_comp_obj_ils = {}

    def finalise_input_source(self, inputListId):

        pass



#!/bin/env python

import sys


def run(config_file, simulator):

    if simulator == "NEURON":
        from bmtk.simulator import bionet

        conf = bionet.Config.from_json(config_file, validate=True)
        conf.build_env()
        net = bionet.BioNetwork.from_config(conf)
        sim = bionet.BioSimulator.from_config(conf, network=net)

    elif simulator == "NEST":
        from bmtk.simulator import pointnet

        conf = pointnet.Config.from_json(config_file)
        conf.build_env()
        net = pointnet.PointNetwork.from_config(conf)
        sim = pointnet.PointSimulator.from_config(conf, net)

    sim.run()


if __name__ == "__main__":

    run("config.json", "NEST")



from neuromllite import Network, Cell, Population, Simulation, Synapse
from neuromllite import RectangularRegion, RandomLayout
from neuromllite import Projection, RandomConnectivity, OneToOneConnector

import sys


def generate(ref="Example6_PyNN", add_inputs=True):

    ################################################################################
    ###   Build new network

    net = Network(id=ref, notes="Another network for PyNN - work in progress...")

    net.parameters = {
        "N_scaling": 0.005,
        "layer_height": 400,
        "width": 100,
        "depth": 100,
        "input_weight": 0.1,
    }

    cell = Cell(id="CorticalCell", pynn_cell="IF_curr_exp")
    cell.parameters = {
        "cm": 0.25,  # nF
        "i_offset": 0.0,  # nA
        "tau_m": 10.0,  # ms
        "tau_refrac": 2.0,  # ms
        "v_reset": -65.0,  # mV
        "v_rest": -65.0,  # mV
        "v_thresh": -50.0,  # mV
    }
    net.cells.append(cell)

    if add_inputs:
        input_cell = Cell(id="InputCell", pynn_cell="SpikeSourcePoisson")
        input_cell.parameters = {"start": 0, "duration": 10000000000, "rate": 150}
        net.cells.append(input_cell)

    e_syn = Synapse(
        id="ampa",
        pynn_receptor_type="excitatory",
        pynn_synapse_type="curr_exp",
        parameters={"tau_syn": 0.5},
    )
    net.synapses.append(e_syn)
    i_syn = Synapse(
        id="gaba",
        pynn_receptor_type="inhibitory",
        pynn_synapse_type="curr_exp",
        parameters={"tau_syn": 0.5},
    )
    net.synapses.append(i_syn)

    N_full = {
        "L23": {"E": 20683, "I": 5834},
        "L4": {"E": 21915, "I": 5479},
        "L5": {"E": 4850, "I": 1065},
        "L6": {"E": 14395, "I": 2948},
    }

    scale = 0.1

    pops = []
    input_pops = []
    pop_dict = {}

    layers = ["L23"]
    layers = ["L23", "L4", "L5", "L6"]

    for l in layers:

        i = 3 - layers.index(l)
        r = RectangularRegion(
            id=l,
            x=0,
            y=i * net.parameters["layer_height"],
            z=0,
            width=net.parameters["width"],
            height=net.parameters["layer_height"],
            depth=net.parameters["depth"],
        )
        net.regions.append(r)

        for t in ["E", "I"]:

            try:
                import opencortex.utils.color as occ

                if l == "L23":
                    if t == "E":
                        color = occ.L23_PRINCIPAL_CELL
                    if t == "I":
                        color = occ.L23_INTERNEURON
                if l == "L4":
                    if t == "E":
                        color = occ.L4_PRINCIPAL_CELL
                    if t == "I":
                        color = occ.L4_INTERNEURON
                if l == "L5":
                    if t == "E":
                        color = occ.L5_PRINCIPAL_CELL
                    if t == "I":
                        color = occ.L5_INTERNEURON
                if l == "L6":
                    if t == "E":
                        color = occ.L6_PRINCIPAL_CELL
                    if t == "I":
                        color = occ.L6_INTERNEURON

            except:
                color = ".8 0 0" if t == "E" else "0 0 1"

            pop_id = "%s_%s" % (l, t)
            pops.append(pop_id)
            ref = "l%s%s" % (l[1:], t.lower())
            exec(
                ref
                + " = Population(id=pop_id, size='int(%s*N_scaling)'%N_full[l][t], component=cell.id, properties={'color':color, 'type':t})"
            )
            exec("%s.random_layout = RandomLayout(region = r.id)" % ref)
            exec("net.populations.append(%s)" % ref)
            exec("pop_dict['%s'] = %s" % (pop_id, ref))

            if add_inputs:
                color = ".8 .8 .8"
                input_id = "%s_%s_input" % (l, t)
                input_pops.append(input_id)
                input_ref = "l%s%s_i" % (l[1:], t.lower())
                exec(
                    input_ref
                    + " = Population(id=input_id, size='int(%s*N_scaling)'%N_full[l][t], component=input_cell.id, properties={'color':color})"
                )
                exec("%s.random_layout = RandomLayout(region = r.id)" % input_ref)
                exec("net.populations.append(%s)" % input_ref)

        # l23i = Population(id='L23_I', size=int(100*scale), component=cell.id, properties={'color':})
        # l23ei = Population(id='L23_E_input', size=int(100*scale), component=input_cell.id)
        # l23ii = Population(id='L23_I_input', size=int(100*scale), component=input_cell.id)

    # net.populations.append(l23e)
    # net.populations.append(l23ei)
    # net.populations.append(l23i)
    # net.populations.append(l23ii)

    conn_probs = [
        [0.1009, 0.1689, 0.0437, 0.0818, 0.0323, 0.0, 0.0076, 0.0],
        [0.1346, 0.1371, 0.0316, 0.0515, 0.0755, 0.0, 0.0042, 0.0],
        [0.0077, 0.0059, 0.0497, 0.135, 0.0067, 0.0003, 0.0453, 0.0],
        [0.0691, 0.0029, 0.0794, 0.1597, 0.0033, 0.0, 0.1057, 0.0],
        [0.1004, 0.0622, 0.0505, 0.0057, 0.0831, 0.3726, 0.0204, 0.0],
        [0.0548, 0.0269, 0.0257, 0.0022, 0.06, 0.3158, 0.0086, 0.0],
        [0.0156, 0.0066, 0.0211, 0.0166, 0.0572, 0.0197, 0.0396, 0.2252],
        [0.0364, 0.001, 0.0034, 0.0005, 0.0277, 0.008, 0.0658, 0.1443],
    ]

    if add_inputs:
        for p in pops:
            proj = Projection(
                id="proj_input_%s" % p,
                presynaptic="%s_input" % p,
                postsynaptic=p,
                synapse=e_syn.id,
                delay=2,
                weight="input_weight",
            )
            proj.one_to_one_connector = OneToOneConnector()
            net.projections.append(proj)

    for pre_i in range(len(pops)):
        for post_i in range(len(pops)):
            pre = pops[pre_i]
            post = pops[post_i]
            prob = conn_probs[post_i][pre_i]  #######   TODO: check!!!!
            weight = 1
            syn = e_syn
            if prob > 0:
                if "I" in pre:
                    weight = -1
                    syn = i_syn
                proj = Projection(
                    id="proj_%s_%s" % (pre, post),
                    presynaptic=pre,
                    postsynaptic=post,
                    synapse=syn.id,
                    delay=1,
                    weight=weight,
                )
                proj.random_connectivity = RandomConnectivity(probability=prob)
                net.projections.append(proj)

    print(net.to_json())
    new_file = net.to_json_file("%s.json" % net.id)

    ################################################################################
    ###   Build Simulation object & save as JSON

    record_traces = {}
    record_spikes = {}

    from neuromllite.utils import evaluate

    for p in pops:
        forecast_size = evaluate(pop_dict[p].size, net.parameters)
        record_traces[p] = list(range(min(2, forecast_size)))
        record_spikes[p] = "*"
    for ip in input_pops:
        record_spikes[ip] = "*"

    sim = Simulation(
        id="Sim%s" % net.id,
        network=new_file,
        duration="100",
        dt="0.025",
        seed=1234,
        record_traces=record_traces,
        record_spikes=record_spikes,
    )

    sim.to_json_file()

    return sim, net


if __name__ == "__main__":

    if "-noinputs" in sys.argv:
        sim, net = generate("Example6_PyNN_noinputs", False)
    else:
        sim, net = generate("Example6_PyNN", True)

    ################################################################################
    ###   Run in some simulators

    from neuromllite.NetworkGenerator import check_to_generate_or_run

    check_to_generate_or_run(sys.argv, sim)



from neuromllite import *
from neuromllite.NetworkGenerator import *
from neuromllite.utils import load_network_json
import sys

################################################################################
###   Reuse network from Example2

filename = "Example2_TestNetwork.json"
net = load_network_json(filename)
net.id = "Example3_Network"
net.notes = "Example 3: simple network with 2 populations of NeuroML2 cells, a projection between them and spiking input."
print(net)
new_file = net.to_json_file()

################################################################################
###   Build Simulation object & save as JSON


sim = Simulation(
    id="SimExample3",
    network=new_file,
    duration="1000",
    dt="0.025",
    record_traces={"all": "*"},
)

sim.to_json_file()


################################################################################
###   Run in some simulators

from neuromllite.NetworkGenerator import check_to_generate_or_run
import sys

check_to_generate_or_run(sys.argv, sim)



from neuromllite import *
from neuromllite.NetworkGenerator import *
from neuromllite.utils import create_new_model
import sys


if __name__ == "__main__":

    hhcell = Cell(
        id="hhcell", neuroml2_source_file="../../examples/test_files/hhcell.cell.nml"
    )

    iclamp = InputSource(
        id="iclamp_0",
        neuroml2_input="PulseGenerator",
        parameters={"amplitude": "stim_amp", "delay": "stim_del", "duration": "200ms"},
    )

    parameters = {"stim_amp": "100pA", "stim_del": "50ms"}

    sim, network = create_new_model(
        "HHTest",
        300,
        parameters=parameters,
        cell_for_default_population=hhcell,
        input_for_default_population=iclamp,
    )

    """
    pynncell = Cell(id='pynncell', 
                    pynn_cell='IF_cond_alpha',
                    parameters = { "tau_refrac":5, "i_offset":0 })

    
    sim, network = create_new_model('PyNNTest',
                     300,
                     parameters = parameters,
                     cell_for_default_population=pynncell,
                     input_for_default_population=iclamp)"""

    check_to_generate_or_run(sys.argv, sim)



import collections

__version__ = "0.5.1"

# import pyNN
# import nest

from typing import List, Dict, Any, Optional, Union, Tuple, Callable, TypeVar, Type

import modelspec
from modelspec import optional, instance_of, in_, has, field, fields
from modelspec.base_types import Base, ValueExprType, value_expr_types


def convert2float(x: Any) -> float:
    """Convert to float if not None"""
    if x is not None:
        return float(x)
    else:
        return None


def convert2int(x: Any) -> float:
    """Convert to int if not None"""
    if x is not None:
        return int(x)
    else:
        return None


class NetworkReaderX:
    pop_locations = {}

    def parse(self, handler):
        raise Exception("This needs to be implemented...")

    def get_locations(self):
        return self.pop_locations


@modelspec.define
class NMLBase(Base):
    """Base class for NeuroML objects."""
    notes: str = field(kw_only=True, default=None, validator=optional(instance_of(str)))


@modelspec.define
class Cell(NMLBase):
    """
    A Cell definition.

    Args:
        id: Unique identifier for this Cell
        parameters: Dictionary of parameters for the cell
        neuroml2_source_file: File name of NeuroML2 file defining the cell
        lems_source_file: File name of LEMS file defining the cell
        neuroml2_cell: Name of standard NeuroML2 cell type
        pynn_cell: Name of standard PyNN cell type
        arbor_cell: Name of standard Arbor cell type
        bindsnet_node: Name of standard BindsNET node
    """
    id: str = field(validator=instance_of(str))
    parameters: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    neuroml2_source_file: str = field(default=None, validator=optional(instance_of(str)))
    lems_source_file: str = field(default=None, validator=optional(instance_of(str)))
    neuroml2_cell: str = field(default=None, validator=optional(instance_of(str)))
    pynn_cell: str = field(default=None, validator=optional(instance_of(str)))
    arbor_cell: str = field(default=None, validator=optional(instance_of(str)))
    bindsnet_node: str = field(default=None, validator=optional(instance_of(str)))


@modelspec.define
class Synapse(NMLBase):
    """
    A Synapse definition.

    Args:
        id: Unique identifier for this Synapse
        parameters: Dictionary of parameters for the synapse
        neuroml2_source_file: File name of NeuroML2 file defining the synapse
        lems_source_file: File name of LEMS file defining the synapse
        pynn_synapse_type: The pynn synapse type. Valid values are: "curr_exp", "curr_alpha", "cond_exp", "cond_alpha".
        pynn_receptor_type: The pynn receptor type. Valid values are: "excitatory", "inhibitory".
    """
    id: str = field(validator=instance_of(str))
    parameters: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    neuroml2_source_file: str = field(default=None, validator=optional(instance_of(str)))
    lems_source_file: str = field(default=None, validator=optional(instance_of(str)))
    pynn_synapse_type: str = field(default=None, validator=optional(in_(["curr_exp", "curr_alpha", "cond_exp", "cond_alpha"])))
    pynn_receptor_type: str = field(default=None, validator=optional(in_(["excitatory", "inhibitory"])))


@modelspec.define
class InputSource(NMLBase):
    """
    An InputSource definition.

    Args:
        id: Unique identifier for this InputSource
        parameters: Dictionary of parameters for the InputSource
        neuroml2_source_file: File name of NeuroML2 file defining the input source
        neuroml2_input: Name of standard NeuroML2 input
        lems_source_file: File name of LEMS file defining the input source
        pynn_input: Name of PyNN input
    """
    id: str = field(validator=instance_of(str))
    parameters: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    neuroml2_source_file: str = field(default=None, validator=optional(instance_of(str)))
    neuroml2_input: str = field(default=None, validator=optional(instance_of(str)))
    lems_source_file: str = field(default=None, validator=optional(instance_of(str)))
    pynn_input: str = field(default=None, validator=optional(instance_of(str)))


@modelspec.define
class RectangularRegion(NMLBase):
    """
    A RectangularRegion definition.

    Args:
        id: Unique identifier for this rectangular region.
        x: x coordinate of corner of region
        y: y coordinate of corner of region
        z: z coordinate of corner of region
        width: width of the rectangular region
        height: height of the rectangular region
        depth: depth of the rectangular region
    """
    id: str = field(validator=instance_of(str))
    x: float = field(validator=instance_of(float), converter=convert2float)
    y: float = field(validator=instance_of(float), converter=convert2float)
    z: float = field(validator=instance_of(float), converter=convert2float)
    width: float = field(validator=instance_of(float), converter=convert2float)
    height: float = field(validator=instance_of(float), converter=convert2float)
    depth: float = field(validator=instance_of(float), converter=convert2float)


@modelspec.define
class RandomLayout(NMLBase):
    """
    A RandomLayout definition.

    Args:
        region: Region in which to place population
    """
    region: str = field(validator=instance_of(str))


@modelspec.define
class RelativeLayout(NMLBase):
    """
    A RelativeLayout definition.

    Args:
        region: The Region relative to which population should be positioned.
        x: x position relative to x coordinate of Region
        y: y position relative to y coordinate of Region
        z: z position relative to z coordinate of Region
    """
    region: str = field(validator=instance_of(str))
    x: float = field(validator=instance_of(float), converter=convert2float)
    y: float = field(validator=instance_of(float), converter=convert2float)
    z: float = field(validator=instance_of(float), converter=convert2float)


@modelspec.define
class Location(NMLBase):
    """
    A Location definition.

    Args:
        x: x coordinate of location
        y: y coordinate of location
        z: z coordinate of location
    """
    x: float = field(validator=instance_of(float), converter=convert2float)
    y: float = field(validator=instance_of(float), converter=convert2float)
    z: float = field(validator=instance_of(float), converter=convert2float)


@modelspec.define
class SingleLocation(NMLBase):
    """
    A SingleLocation definition.

    Args:
        location: Location of the single Cell.
    """
    location: Location = field(validator=instance_of(Location))


@modelspec.define
class Population(NMLBase):
    """
    A Population definition.

    Args:
        id: Unique identifier for this Population
        size: The size of the population.
        component: The type of Cell to use in this population.
        properties: A dictionary of properties (metadata) for this population.
        random_layout: Layout in the random RectangularRegion.
        relative_layout: Position relative to RectangularRegion.
        single_location: Explicit location of the one Cell in the population
    """
    id: str = field(validator=instance_of(str))
    size: ValueExprType = field(validator=instance_of(value_expr_types))
    component: str = field(validator=instance_of(str))
    properties: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    random_layout: RandomLayout = field(default=None, validator=optional(instance_of(RandomLayout)))
    relative_layout: RelativeLayout = field(default=None, validator=optional(instance_of(RelativeLayout)))
    single_location: SingleLocation = field(default=None, validator=optional(instance_of(SingleLocation)))

    def has_positions(self):
        """
        Returns True if the population has a position.

        Returns:
            True if the population has a position.
        """
        if self.random_layout is not None:
            return True
        elif self.relative_layout is not None:
            return True
        elif self.single_location is not None:
            return True
        else:
            return False


@modelspec.define
class RandomConnectivity(NMLBase):
    """
    A RandomConnectivity definition.

    Args:
        probability: Random probability of connection.
    """
    probability: ValueExprType = field(validator=instance_of(value_expr_types))


@modelspec.define
class OneToOneConnector(NMLBase):
    """
    A OneToOneConnector definition.
    """
    pass


# Temp! to redefine more generally!
@modelspec.define
class ConvergentConnectivity(NMLBase):
    """
    A ConvergentConnectivity definition.

    Args:
        num_per_post: Number per post-synaptic neuron.
    """
    num_per_post: float = field(validator=instance_of(float), converter=convert2float)


@modelspec.define
class Projection(NMLBase):
    """
    A Projection definition.

    Args:
        id: Unique identifier for this Projection
        presynaptic: Presynaptic Population
        postsynaptic: Postsynaptic Population
        synapse: Which Synapse to use
        pre_synapse: For continuous connections, what presynaptic component to use (default: silent analog synapse)
        type: type of projection: projection (default; standard chemical, event triggered), electricalProjection
            (for gap junctions) or continuousProjection (for analogue/graded synapses)
        delay: Delay to use (default: 0)
        weight: Weight to use (default: 1)
        random_connectivity: Use random connectivity
        convergent_connectivity: Use convergent connectivity
        one_to_one_connector: Connect cell index i in pre pop to cell index i in post pop for all i

    """
    id: str = field(validator=instance_of(str))
    presynaptic: str = field(validator=optional(instance_of(str)))
    postsynaptic: str = field(validator=optional(instance_of(str)))
    synapse: str = field(validator=optional(instance_of(str)))
    pre_synapse: str = field(default=None, validator=optional(instance_of(str)))
    type: str = field(default="projection", validator=optional(instance_of(str)))
    delay: ValueExprType = field(default=None, validator=optional(instance_of(value_expr_types)))
    weight: ValueExprType = field(default=None, validator=optional(instance_of(value_expr_types)))
    random_connectivity: RandomConnectivity = field(default=None, validator=optional(instance_of(RandomConnectivity)))
    convergent_connectivity: ConvergentConnectivity = field(default=None, validator=optional(instance_of(ConvergentConnectivity)))
    one_to_one_connector: OneToOneConnector = field(default=None, validator=optional(instance_of(OneToOneConnector)))


@modelspec.define
class Input(NMLBase):
    """
    An Input definition.

    Args:
        id: Unique identifier for this Input
        input_source: Type of input to use in population
        population: Population to target
        percentage: Percentage of Cells to apply input to
        cell_ids: Specific ids of _Cell_s to apply this input to (cannot be used with percentage
        number_per_cell: Number of individual inputs per selected Cell (default: 1)
        segment_ids: Which segments to target (default: [0])
        weight: Weight to use (default: 1)
    """

    id: str = field(validator=instance_of(str))
    input_source: str = field(default=None, validator=optional(instance_of(str)))
    population: str = field(default=None, validator=optional(instance_of(str)))
    cell_ids: ValueExprType = field(default="", validator=optional(instance_of(value_expr_types)))
    percentage: float = field(default=None, validator=optional(instance_of(float)), converter=convert2float)
    number_per_cell: ValueExprType = field(default="", validator=optional(instance_of(value_expr_types)))
    segment_ids: ValueExprType = field(default="", validator=optional(instance_of(value_expr_types)))
    weight: ValueExprType = field(default=None, validator=optional(instance_of(value_expr_types)))


@modelspec.define
class NetworkReader(NMLBase):
    """
    A NetworkReader definition.

    Args:
        type: The type of NetworkReader
        parameters: Dictionary of parameters for the NetworkReader
    """
    type: str = field(default=None, validator=optional(instance_of(str)))
    parameters: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))


@modelspec.define
class Simulation(NMLBase):
    """
    A Simulation definition.

    Args:
        id: Unique identifier for this Simulation
        version: Information on verson of NeuroMLlite
        network: File name of network to simulate
        duration: Duration of simulation (ms)
        dt: Timestep of simulation (ms)
        seed: Seed for stochastic elements os the simulation
        record_traces: Record traces?
        record_spikes: Record spikes?
        record_rates: Record rates?
        record_variables: Record named variables?
        plots2D: Work in progress...
        plots3D: Work in progress...
    """
    id: str = field(validator=instance_of(str))
    version: str = field(default=f"NeuroMLlite v{__version__}", validator=optional(instance_of(str)), metadata={"omit_if_default": False})
    network: str = field(default=None, validator=optional(instance_of(str)))
    duration: float = field(default=None, validator=optional(instance_of(float)), converter=convert2float)
    dt: float = field(default=None, validator=optional(instance_of(float)), converter=convert2float)
    seed: int = field(default=None, validator=optional(instance_of(int)), converter=convert2int)
    record_traces: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    record_spikes: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    record_rates: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    record_variables: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    plots2D: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    plots3D: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))


@modelspec.define
class Network(NMLBase):
    """
    A Network containing multiple Population's, connected by Projection's and receiving Input's

    Args:
        id: Unique identifier for the Network
        parameters: Dictionary of global parameters for the network
        cells: The Cells which can be present in Populations
        synapses: The Synapse definitions which are used in Projections
        input_sources: The InputSource definitions which define the types of stimulus which can be applied in Inputs
        regions: The Regions in which Populations get placed.
        populations: The Populations of Cells making up this network ...
        projections: The Projections between Populations
        inputs: The inputs to apply to the elements of Populations
        version: Information on verson of NeuroMLlite
        seed: Seed for random number generator used when building network
        temperature: Temperature at which to run network (float in deg C)
        network_reader: A class which can read in a network (e.g. from a structured format)
        notes: Human readable notes about the network
    """

    id: str = field(validator=instance_of(str))
    parameters: Dict[str, Any] = field(default=None, validator=optional(instance_of(dict)))
    cells: List[Cell] = field(factory=list, validator=instance_of(list))
    synapses: List[Synapse] = field(factory=list, validator=instance_of(list))
    input_sources: List[InputSource] = field(factory=list, validator=instance_of(list))
    regions: List[RectangularRegion] = field(factory=list, validator=instance_of(list))
    populations: List[Population] = field(factory=list, validator=instance_of(list))
    projections: List[Projection] = field(factory=list, validator=instance_of(list))
    inputs: List[Input] = field(factory=list, validator=instance_of(list))
    version: str = field(default=f"NeuroMLlite v{__version__}", validator=instance_of(str), metadata={"omit_if_default": False})
    seed: int = field(default=None, validator=optional(instance_of(int)))
    temperature: float = field(default=None, validator=optional(instance_of(float)), converter=convert2float)
    network_reader: NetworkReader = field(default=None)


if __name__ == "__main__":

    net = Network(id="net")
    doc = net.generate_documentation(format="markdown")
    print(doc)
    with open("../docs/README.md", "w") as d:
        d.write(doc)

    import json
    import yaml

    doc = net.generate_documentation(format="dict")
    doc = {"version": "NeuroMLlite v%s" % __version__, "specification": doc}
    with open("../docs/NeuroMLlite_specification.json", "w") as d:
        d.write(json.dumps(doc, indent=4))
    with open("../docs/NeuroMLlite_specification.yaml", "w") as d:
        d.write(yaml.dump(doc, indent=4, sort_keys=False))



from neuromllite import Cell
from neuromllite import Input
from neuromllite import InputSource
from neuromllite import Network
from neuromllite import Population
from neuromllite import RandomLayout
from neuromllite import RectangularRegion
from neuromllite import Simulation
from neuromllite import Synapse
from neuromllite import Projection
from neuromllite import RandomConnectivity

import sys

################################################################################
###   Build new network

net = Network(id="ArborExample")
net.notes = "Example for testing Arbor"

net.parameters = {
    "v_init": -50,
    "scale": 3,
    "input_amp": 0.01,
    "input_del": 50,
    "input_dur": 5,
}

cell = Cell(id="test_arbor_cell", arbor_cell="cable_cell")

cell.parameters = {"v_init": "v_init", "radius": 3, "mechanism": "hh"}


net.cells.append(cell)
"""
cell2 = Cell(id='testcell2', pynn_cell='IF_cond_alpha')
cell2.parameters = { "tau_refrac":5, "i_offset":-.1 }
net.cells.append(cell2)"""

input_source = InputSource(
    id="i_clamp",
    pynn_input="DCSource",
    parameters={
        "amplitude": "input_amp",
        "start": "input_del",
        "stop": "input_del+input_dur",
    },
)

net.input_sources.append(input_source)

r0 = RectangularRegion(id="region0", x=0, y=0, z=0, width=1000, height=100, depth=1000)
net.regions.append(r0)
r1 = RectangularRegion(
    id="region1", x=0, y=200, z=0, width=1000, height=100, depth=1000
)
net.regions.append(r1)

p0 = Population(
    id="pop0",
    size="scale",
    component=cell.id,
    properties={"color": "1 0 0"},
    random_layout=RandomLayout(region=r0.id),
)
net.populations.append(p0)
p1 = Population(
    id="pop1",
    size="scale",
    component=cell.id,
    properties={"color": "0 1 0"},
    random_layout=RandomLayout(region=r1.id),
)
net.populations.append(p1)
"""
p1 = Population(id='pop1', size=2, component=cell2.id, properties={'color':'0 1 0'},random_layout = RandomLayout(region=r1.id))
p2 = Population(id='pop2', size=1, component=cell2.id, properties={'color':'0 0 1'},random_layout = RandomLayout(region=r1.id))

net.populations.append(p1)
net.populations.append(p2)"""


net.synapses.append(
    Synapse(
        id="ampaSyn",
        pynn_receptor_type="excitatory",
        pynn_synapse_type="cond_alpha",
        parameters={"e_rev": -10, "tau_syn": 2},
    )
)


net.projections.append(
    Projection(
        id="proj0",
        presynaptic=p0.id,
        postsynaptic=p1.id,
        synapse="ampaSyn",
        delay="5",
        weight="0.0001*random()",
    )
)

net.projections[0].random_connectivity = RandomConnectivity(probability=0.5)

'''
net.synapses.append(Synapse(id='gabaSyn',
                            pynn_receptor_type='inhibitory',
                            pynn_synapse_type='cond_alpha',
                            parameters={'e_rev':-80, 'tau_syn':10}))


net.projections.append(Projection(id='proj1',
                                  presynaptic=p0.id,
                                  postsynaptic=p2.id,
                                  synapse='gabaSyn',
                                  delay=2,
                                  weight=0.01))
net.projections[1].random_connectivity=RandomConnectivity(probability=1)'''

net.inputs.append(
    Input(id="stim", input_source=input_source.id, population=p0.id, cell_ids=[1,2])
)

print(net.to_json())
new_file = net.to_json_file("%s.json" % net.id)
new_file_yaml = net.to_yaml_file("%s.yaml" % net.id)


################################################################################
###   Build Simulation object & save as JSON

sim = Simulation(
    id="Sim%s" % net.id,
    network=new_file,
    duration="100",
    dt="0.01",
    record_traces={"all": "*"},
    record_spikes={"pop0": "*"},
)

sim.to_json_file()
sim.network = new_file_yaml
sim.to_yaml_file('%s.yaml'%sim.id)


################################################################################
###   Run in some simulators

from neuromllite.NetworkGenerator import check_to_generate_or_run
import sys

check_to_generate_or_run(sys.argv, sim)



import tables  # pytables for HDF5 support
import os

from neuroml.hdf5.NetworkContainer import *
from neuromllite import NetworkReaderX

from neuromllite.utils import print_v


class BBPConnectomeReader(NetworkReaderX):

    component_objects = {}  # Store cell ids vs objects, e.g. NeuroML2 based object

    def __init__(self, **parameters):

        print_v("Creating BBPConnectomeReader with %s..." % parameters)
        self.parameters = parameters
        self.current_population = None
        self.pre_pop = None
        self.post_pop = None

    def parse(self, handler):

        filename = os.path.abspath(self.parameters["filename"])
        id = filename.split("/")[-1].split(".")[0]

        if "id" in self.parameters:
            id = self.parameters["id"]

        self.handler = handler

        notes = "Network read in from BBP connectome: %s" % filename
        handler.handle_document_start(id, notes)

        handler.handle_network(id, notes)

        h5file = tables.open_file(filename, mode="r")

        print_v("Opened HDF5 file: %s" % (h5file.filename))

        self.parse_group(h5file.root.populations)
        self.parse_group(h5file.root.connectivity)

        h5file.close()

    def parse_group(self, g):
        # print("+++++++++++++++Parsing group: "+ str(g)+", name: "+g._v_name)

        for node in g:
            # print("   ------Sub node: %s, class: %s, name: %s (parent: %s)"   % (node,node._c_classid,node._v_name, g._v_name))

            if node._c_classid == "GROUP":
                if g._v_name == "populations":
                    pop_id = node._v_name.replace("-", "_")
                    self.current_population = pop_id
                    self.pop_locations[self.current_population] = {}

                if g._v_name == "connectivity":
                    self.pre_pop = node._v_name.replace("-", "_")
                    self.post_pop = None
                    # print("Conn %s -> ?"%(self.pre_pop))
                elif self.pre_pop != None and self.post_pop == None:
                    self.post_pop = node._v_name.replace("-", "_")
                    # print("Conn2 %s -> %s"%(self.pre_pop,self.post_pop))
                elif self.pre_pop != None and self.post_pop != None:
                    # print("Conn3 %s -> %s"%(self.pre_pop,self.post_pop))
                    pass

                self.parse_group(node)

            if self._is_dataset(node):
                self.parse_dataset(node)

        self.current_population = None

    def _is_dataset(self, node):
        return node._c_classid == "ARRAY" or node._c_classid == "CARRAY"

    def _is_interneuron(self, pop_id):
        if "PC" in pop_id or "SS" in pop_id or "SP" in pop_id:
            return False
        else:
            return True

    def parse_dataset(self, d):
        # print_v("Parsing dataset/array: "+ str(d))

        # Population
        if self.current_population and d.name == "locations":

            perc_cells = (
                self.parameters["percentage_cells_per_pop"]
                if "percentage_cells_per_pop" in self.parameters
                else 100
            )
            if perc_cells > 100:
                perc_cells = 100

            size = max(0, int((perc_cells / 100.0) * d.shape[0]))

            if size > 0:
                properties = {}
                if self._is_interneuron(self.current_population):
                    properties["radius"] = 5
                    type = "I"
                else:
                    properties["radius"] = 10
                    type = "E"
                properties["type"] = type

                layer = self.current_population.split("_")[0]
                properties["region"] = layer
                try:
                    import opencortex.utils.color as occ

                    if layer == "L23":
                        if type == "E":
                            color = occ.L23_PRINCIPAL_CELL
                        if type == "I":
                            color = occ.L23_INTERNEURON
                    if layer == "L4":
                        if type == "E":
                            color = occ.L4_PRINCIPAL_CELL
                        if type == "I":
                            color = occ.L4_INTERNEURON
                    if layer == "L5":
                        if type == "E":
                            color = occ.L5_PRINCIPAL_CELL
                        if type == "I":
                            color = occ.L5_INTERNEURON
                    if layer == "L6":
                        if type == "E":
                            color = occ.L6_PRINCIPAL_CELL
                        if type == "I":
                            color = occ.L6_INTERNEURON

                    properties["color"] = color
                except:
                    # Don't worry about it, it's just metadata
                    pass

                component_obj = None

                if self.parameters["DEFAULT_CELL_ID"] in self.component_objects:
                    component_obj = self.component_objects[
                        self.parameters["DEFAULT_CELL_ID"]
                    ]
                else:
                    if "cell_info" in self.parameters:
                        def_cell_info = self.parameters["cell_info"][
                            self.parameters["DEFAULT_CELL_ID"]
                        ]
                        if def_cell_info.neuroml2_source_file:
                            from pyneuroml import pynml

                            nml2_doc = pynml.read_neuroml2_file(
                                def_cell_info.neuroml2_source_file,
                                include_includes=True,
                            )
                            component_obj = nml2_doc.get_by_id(
                                self.parameters["DEFAULT_CELL_ID"]
                            )
                            print_v(
                                "Loaded NeuroML2 object %s from %s "
                                % (component_obj, def_cell_info.neuroml2_source_file)
                            )
                            self.component_objects[
                                self.parameters["DEFAULT_CELL_ID"]
                            ] = component_obj

                self.handler.handle_population(
                    self.current_population,
                    self.parameters["DEFAULT_CELL_ID"],
                    size,
                    component_obj=component_obj,
                    properties=properties,
                )

                print_v(
                    "   There are %i cells in: %s" % (size, self.current_population)
                )
                for i in range(0, d.shape[0]):

                    if i < size:
                        row = d[i, :]
                        x = row[0]
                        y = row[1]
                        z = row[2]
                        self.pop_locations[self.current_population][i] = (x, y, z)
                        self.handler.handle_location(
                            i,
                            self.current_population,
                            self.parameters["DEFAULT_CELL_ID"],
                            x,
                            y,
                            z,
                        )

        # Projection
        elif self.pre_pop != None and self.post_pop != None:

            proj_id = "Proj__%s__%s" % (self.pre_pop, self.post_pop)
            synapse = "gaba"
            if (
                "PC" in self.pre_pop or "SS" in self.pre_pop
            ):  # TODO: better choice between E/I cells
                synapse = "ampa"

            (ii, jj) = np.nonzero(d)
            conns_here = False
            pre_num = (
                len(self.pop_locations[self.pre_pop])
                if self.pre_pop in self.pop_locations
                else 0
            )
            post_num = (
                len(self.pop_locations[self.post_pop])
                if self.post_pop in self.pop_locations
                else 0
            )

            if pre_num > 0 and post_num > 0:
                for index in range(len(ii)):
                    if ii[index] < pre_num and jj[index] < post_num:
                        conns_here = True
                        break

                if conns_here:
                    print_v(
                        "Conn %s -> %s (%s)" % (self.pre_pop, self.post_pop, synapse)
                    )
                    self.handler.handle_projection(
                        proj_id, self.pre_pop, self.post_pop, synapse
                    )

                    conn_count = 0

                    for index in range(len(ii)):
                        i = ii[index]
                        j = jj[index]
                        if i < pre_num and j < post_num:
                            # print("  Conn5 %s[%s] -> %s[%s]"%(self.pre_pop,i,self.post_pop,j))
                            delay = 1.111
                            weight = 1
                            self.handler.handle_connection(
                                proj_id,
                                conn_count,
                                self.pre_pop,
                                self.post_pop,
                                synapse,
                                i,
                                j,
                                delay=delay,
                                weight=weight,
                            )
                            conn_count += 1

                    self.handler.finalise_projection(
                        proj_id, self.pre_pop, self.post_pop, synapse
                    )

            self.post_pop = None


if __name__ == "__main__":

    filename = "test_files/cons_locs_pathways_mc0_Column.h5"

    percentage_cells_per_pop = 1

    bbp = BBPConnectomeReader(
        filename=filename,
        percentage_cells_per_pop=percentage_cells_per_pop,
        DEFAULT_CELL_ID="hhcell",
    )

    from neuromllite.DefaultNetworkHandler import DefaultNetworkHandler

    def_handler = DefaultNetworkHandler()

    bbp.parse(def_handler)

    from neuroml.hdf5.NetworkBuilder import NetworkBuilder

    neuroml_handler = NetworkBuilder()

    bbp = BBPConnectomeReader(
        filename=filename,
        percentage_cells_per_pop=percentage_cells_per_pop,
        DEFAULT_CELL_ID="hhcell",
    )
    bbp.parse(neuroml_handler)

    nml_file_name = "BBP.net.nml"

    from neuroml.writers import NeuroMLWriter

    NeuroMLWriter.write(neuroml_handler.get_nml_doc(), nml_file_name)

    """
    from neuromllite.NeuronHandler import NeuronHandler


    nrn_handler = NeuronHandler()

    nrn_handler.executeHoc('load_file("hhcell.hoc")')
    bbp = BBPConnectomeReader()
    bbp.parse(file_name, nrn_handler)

    #bbp.save_to_hdf5(nml_h5_file_name)"""



#
#
#   A class to write to the PsyNeuLink/MDF format...
#
#

from neuromllite.DefaultNetworkHandler import DefaultNetworkHandler
from neuromllite.utils import print_v
from neuromllite.utils import save_to_json_file
from neuromllite.utils import locate_file
from neuromllite.utils import evaluate

import lems.api as lems  # from pylems


class PsyNeuLinkHandler(DefaultNetworkHandler):

    bids_mdf_info = {}
    bids_mdf_info["graphs"] = []

    bids_mdf_info_hl = {}
    bids_mdf_info_hl["graphs"] = []

    def __init__(self, nl_network):
        print_v("Initiating PsyNeuLink handler")
        self.nl_network = nl_network

    def handle_document_start(self, id, notes):

        print_v("Parsing for PsyNeuLink export: %s" % id)
        self.id = id

    def finalise_document(self):

        print_v("Writing file for...: %s" % self.id)

        save_to_json_file(self.bids_mdf_info, "%s.bids-mdf.json" % self.id, indent=4)
        """save_to_json_file(self.bids_mdf_info_hl, '%s.bids-mdf.highlevel.json'%self.id, indent=4)"""

    def handle_network(self, network_id, notes, temperature=None):

        print_v("Network: %s" % network_id)
        self.network_id = network_id

        if temperature:
            print_v("  Temperature: " + temperature)
        if notes:
            print_v("  Notes: " + notes)

        self.bids_mdf_graph = {}
        self.bids_mdf_graph["name"] = network_id
        self.bids_mdf_graph["notes"] = notes
        self.bids_mdf_graph["controller"] = None
        self.bids_mdf_graph["nodes"] = {}
        self.bids_mdf_graph["edges"] = {}
        self.bids_mdf_graph["parameters"] = {}
        self.bids_mdf_graph["parameters"]["PNL"] = {"required_node_roles": []}
        self.bids_mdf_info["graphs"].append(self.bids_mdf_graph)
        self.bids_mdf_graph["type"] = {"PNL": "Composition", "generic": "graph"}

        self.bids_mdf_graph_hl = {}
        self.bids_mdf_graph_hl["name"] = network_id + "_hl"
        self.bids_mdf_graph_hl["notes"] = (
            notes + " High Level (population/projection based) description"
        )
        self.bids_mdf_graph_hl["nodes"] = {}
        self.bids_mdf_graph_hl["edges"] = {}
        self.bids_mdf_graph_hl["parameters"] = {}
        self.bids_mdf_info_hl["graphs"].append(self.bids_mdf_graph_hl)

    def handle_population(
        self, population_id, component, size=-1, component_obj=None, properties={}
    ):

        sizeInfo = " as yet unspecified size"
        if size >= 0:
            sizeInfo = ", size: " + str(size) + " cells"
        if component_obj:
            compInfo = " (%s)" % component_obj.__class__.__name__
        else:
            compInfo = ""

        print_v(
            "Population: "
            + population_id
            + ", component: "
            + component
            + compInfo
            + sizeInfo
        )

        if size >= 0:
            for i in range(size):
                node_id = "%s_%i" % (population_id, i)
                node = {}
                node["type"] = {}
                node["name"] = node_id
                # node['type']['NeuroML'] = component

                comp = self.nl_network.get_child(component, "cells")
                base_dir = "./"  # for now...
                fname = locate_file(comp.lems_source_file, base_dir)
                model = lems.Model()
                model.import_from_file(fname)
                lems_comp = model.components.get(component)
                print(
                    "Cell: [%s] comes from %s and in Lems is: %s"
                    % (comp, fname, lems_comp)
                )
                comp_type = lems_comp.type

                type = "The type of %s" % comp_type
                function = "The function of %s" % comp_type

                if comp_type == "pnlLinearFunctionTM":
                    function = "Linear"
                    type = "TransferMechanism"
                elif comp_type == "inputNode":
                    function = "Linear"
                    type = "TransferMechanism"
                elif comp_type == "pnlLogisticFunctionTM":
                    function = "Logistic"
                    type = "TransferMechanism"
                elif comp_type == "pnlExponentialFunctionTM":
                    function = "Exponential"
                    type = "TransferMechanism"
                elif comp_type == "pnlSimpleIntegratorMechanism":
                    function = "SimpleIntegrator"
                    type = "IntegratorMechanism"
                elif comp_type == "fnCell":
                    function = "FitzHughNagumoIntegrator"
                    type = "IntegratorMechanism"

                node["type"]["PNL"] = type
                node["type"]["generic"] = None
                node["parameters"] = {}
                node["parameters"]["PNL"] = {}
                node["functions"] = []
                func_info = {}
                func_info["type"] = {}
                func_info["type"]["generic"] = function
                func_info["name"] = "Function_%s" % function
                func_info["args"] = {}
                for p in lems_comp.parameters:
                    func_info["args"][p] = {}
                    func_info["args"][p]["type"] = "float"
                    func_info["args"][p]["source"] = "%s.input_ports.%s" % (node_id, p)

                    if comp.parameters is not None and p in comp.parameters:
                        func_info["args"][p]["value"] = evaluate(
                            comp.parameters[p], self.nl_network.parameters
                        )
                    else:
                        func_info["args"][p]["value"] = evaluate(
                            lems_comp.parameters[p]
                        )  # evaluate to ensure strings -> ints/floats etc

                node["functions"].append(func_info)
                self.bids_mdf_graph["nodes"][node_id] = node

            pop_node_id = "%s" % (population_id)
            pop_node = {}
            pop_node["type"] = {}
            pop_node["name"] = pop_node_id
            pop_node["type"]["NeuroML"] = component
            pop_node["parameters"] = {}
            pop_node["parameters"]["size"] = size
            pop_node["functions"] = {}
            self.bids_mdf_graph_hl["nodes"][pop_node_id] = pop_node

    def handle_location(self, id, population_id, component, x, y, z):
        pass

    def finalise_population(self, population_id):

        pass

    #
    #  Should be overridden to handle network connection
    #
    def handle_connection(
        self,
        projName,
        id,
        prePop,
        postPop,
        synapseType,
        preCellId,
        postCellId,
        preSegId=0,
        preFract=0.5,
        postSegId=0,
        postFract=0.5,
        delay=0,
        weight=1,
    ):

        # self.print_connection_information(projName, id, prePop, postPop, synapseType, preCellId, postCellId, weight)
        print_v(
            ">>>>>> Src cell: %d, seg: %f, fract: %f -> Tgt cell %d, seg: %f, fract: %f; weight %s, delay: %s ms"
            % (
                preCellId,
                preSegId,
                preFract,
                postCellId,
                postSegId,
                postFract,
                weight,
                delay,
            )
        )

        pre_node_id = "%s_%i" % (prePop, preCellId)
        post_node_id = "%s_%i" % (postPop, postCellId)
        edge_id = "Edge %s to %s" % (pre_node_id, post_node_id)
        edge = {}
        edge["name"] = edge_id
        edge["type"] = {}
        edge["type"]["NeuroML"] = synapseType
        edge["parameters"] = {}
        edge["functions"] = {}
        edge["sender_port"] = "OutputPort"
        edge["receiver_port"] = "InputPort"
        edge["sender"] = pre_node_id
        edge["receiver"] = post_node_id
        edge["weight"] = weight

        edge["type"] = {"PNL": "MappingProjection", "generic": None}

        self.bids_mdf_graph["edges"][edge_id] = edge

    #
    #  Should be overridden to create input source array
    #
    def handle_input_list(
        self, inputListId, population_id, component, size, input_comp_obj=None
    ):

        pass

    #
    #  Should be overridden to to connect each input to the target cell
    #
    def handle_single_input(
        self, inputListId, id, cellId, segId=0, fract=0.5, weight=1
    ):

        print_v(
            "Input: %s[%s], cellId: %i, seg: %i, fract: %f, weight: %f"
            % (inputListId, id, cellId, segId, fract, weight)
        )

    #
    #  Should be overridden to to connect each input to the target cell
    #
    def finalise_input_source(self, inputName):
        print_v("Input : %s completed" % inputName)
